/*
 * Author:  Ursu Vlad-Andrei
 * Project: Proiect PM 2015-2016 Sudoku
 * File:    sudoku.c
 */

#include <avr/io.h>
#include <util/delay.h>
#include <stdlib.h>
#include <string.h>
#include <avr/interrupt.h>

#include "u8g.h"

#define F_CPU 16000000

#define LCD_WIDTH 128
#define LCD_HEIGHT 64
#define OK 0
#define NOT_OK 1
#define BOX_HEIGHT 6
#define BOX_WIDTH 8
#define TABLE_WIDTH 82
#define TABLE_HEIGHT (LCD_HEIGHT)
#define MENU_CURSOR 26
#define UNIT 10
#define DELAY 500

#define SIZE 9

typedef struct {
    int8_t x;
    int8_t y;
} coord;

/* The states the game can be in */
enum state {UNLOCKED, LOCKED, SELECT_LEVEL};

/* LCD */
static u8g_t u8g;

/* Digit strings */
static char *digits[10] = {"0", "1", "2", "3", "4", "5", "6", "7", "8", "9"};

/* Current cursor position */
static coord current_pos;


/* EASY puzzle */
static uint8_t easy[SIZE * SIZE] = {0, 7, 1, 0, 9, 0, 8, 0, 0,
                                    0, 0, 0, 3, 0, 6, 0, 0, 0,
                                    4, 9, 0, 0, 0, 0, 7, 0, 5,
                                    0, 1, 0, 9, 0, 0, 0, 0, 0,
                                    9, 0, 2, 0, 0, 0, 6, 0, 3,
                                    0, 0, 0, 0, 0, 8, 0, 2, 0,
                                    8, 0, 5, 0, 0, 0, 0, 7, 6,
                                    0, 0, 0, 6, 0, 7, 0, 0, 0,
                                    0, 0, 7, 0, 4, 0, 3, 5, 0};

/* Already completed positions */
static uint8_t easy_completed = 28;

static uint8_t easy_fixed[SIZE * SIZE] = {0, 7, 1, 0, 9, 0, 8, 0, 0,
                                          0, 0, 0, 3, 0, 6, 0, 0, 0,
                                    	  4, 9, 0, 0, 0, 0, 7, 0, 5,
                                    	  0, 1, 0, 9, 0, 0, 0, 0, 0,
                                    	  9, 0, 2, 0, 0, 0, 6, 0, 3,
                                    	  0, 0, 0, 0, 0, 8, 0, 2, 0,
                                    	  8, 0, 5, 0, 0, 0, 0, 7, 6,
                                    	  0, 0, 0, 6, 0, 7, 0, 0, 0,
                                    	  0, 0, 7, 0, 4, 0, 3, 5, 0};



/* MEDIUM puzzle */
static uint8_t medium[SIZE * SIZE] = {6, 0, 2, 0, 9, 1, 0, 0, 0,
                                      3, 5, 0, 8, 2, 0, 0, 0, 0,
                                      0, 0, 0, 0, 0, 0, 1, 0, 2,
                                      0, 8, 0, 0, 4, 0, 0, 0, 9,
                                      0, 0, 4, 0, 8, 0, 7, 0, 0,
                                      2, 0, 0, 0, 7, 0, 0, 1, 0,
                                      4, 0, 9, 0, 0, 0, 0, 0, 0,
                                      0, 0, 0, 0, 5, 4, 0, 3, 6,
                                      0, 0, 0, 7, 1, 0, 2, 0, 4};

static uint8_t medium_fixed[SIZE * SIZE] = {6, 0, 2, 0, 9, 1, 0, 0, 0,
                                            3, 5, 0, 8, 2, 0, 0, 0, 0,
                                            0, 0, 0, 0, 0, 0, 1, 0, 2,
                                            0, 8, 0, 0, 4, 0, 0, 0, 9,
                                            0, 0, 4, 0, 8, 0, 7, 0, 0,
                                            2, 0, 0, 0, 7, 0, 0, 1, 0,
                                            4, 0, 9, 0, 0, 0, 0, 0, 0,
                                            0, 0, 0, 0, 5, 4, 0, 3, 6,
                                            0, 0, 0, 7, 1, 0, 2, 0, 4};


/* Already completed positions */
static uint8_t medium_completed = 29;                                     
                                      

/* HARD puzzle */
static uint8_t hard[SIZE * SIZE] = {0, 7, 5, 0, 0, 0, 0, 2, 0,
                                    8, 0, 0, 0, 2, 0, 9, 1, 0,
                                    4, 0, 0, 0, 0, 0, 0, 0, 0,
                                    0, 0, 0, 0, 0, 1, 0, 2, 6,
                                    9, 0, 0, 2, 0, 5, 0, 0, 3,
                                    7, 3, 0, 8, 0, 0, 0, 0, 0,
                                    0, 0, 0, 0, 0, 0, 0, 0, 8,
                                    0, 5, 4, 0, 2, 0, 0, 0, 7,
                                    0, 8, 0, 0, 0, 0, 3, 4, 0};

static uint8_t hard_fixed[SIZE *SIZE] = {0, 7, 5, 0, 0, 0, 0, 2, 0,
                                         8, 0, 0, 0, 2, 0, 9, 1, 0,
                                         4, 0, 0, 0, 0, 0, 0, 0, 0,
                                         0, 0, 0, 0, 0, 1, 0, 2, 6,
                                         9, 0, 0, 2, 0, 5, 0, 0, 3,
                                         7, 3, 0, 8, 0, 0, 0, 0, 0,
                                         0, 0, 0, 0, 0, 0, 0, 0, 8,
                                         0, 5, 4, 0, 2, 0, 0, 0, 7,
                                         0, 8, 0, 0, 0, 0, 3, 4, 0};

/* Already completed positions */
static uint8_t hard_completed = 26;                                     

/* Current puzzle */
static uint8_t *puzzle;
static uint8_t *fixed_puzzle;
static uint8_t puzzle_completed;

/* Current state */
static enum state current_state;

/* Menu cursor */
static uint8_t menu_cursor = MENU_CURSOR;

/* Initialize LCD */
void init_lcd(void) {

    MCUCR |= (1 << JTD);
    MCUCR |= (1 << JTD);

    /* u8g_II_api.c */
    u8g_Init8Bit(&u8g, &u8g_dev_st7920_128x64_4x_8bit,
        PN(0, 0), PN(0, 1), PN(0, 2), PN(0, 3), 
        PN(0, 4), PN(0, 5), PN(0, 6), PN(0, 7), // DATA - Port A
        PN(3, 2),                               // ENABLE - PD2
        U8G_PIN_NONE,
        U8G_PIN_NONE,
        PN(3, 0),                               // RS/DI - PD0
        PN(3, 1),                               // RW    - PD1
        PN(3, 4));                              // RESET - PD4
}

/* Initialize buttons */
void init_buttons(void) {

    DDRC = 0xff;
    PORTC = 0x00;

    /* LEFT */
    DDRC &= ~(1 << PC0);
    PORTC |= (1 << PC0);
    
    /* UP */
    DDRC &= ~(1 << PC2);
    PORTC |= (1 << PC2);

    /* DOWN */
    DDRC &= ~(1 << PC1);
    PORTC |= (1 << PC1);
    
    /* RIGHT */
    DDRC &= ~(1 << PC3);
    PORTC |= (1 << PC3);

    /* SELECT */
    DDRC &= ~(1 << PC4);
    PORTC |= (1 << PC4);

    /* LED user */
    DDRD |= (1 << PD7);

    /* Enable interrupts  */
    PCICR  = (1 << PCIE2);
    PCMSK2 = (1 << PCINT16) | (1 << PCINT17) | (1 << PCINT18) | (1 << PCINT19)
        | (1 << PCINT20) | (1 << PCINT21);
}

/* Initialize the game */
void init_game(void) {
    current_pos.x = 0;
    current_pos.y = 0;
    current_state = SELECT_LEVEL;
}

/* Draw the select level menu */
void draw_level_select(void) {

    u8g_SetDefaultForegroundColor(&u8g);
    u8g_SetFont(&u8g, u8g_font_6x10);
    u8g_DrawStr(&u8g, UNIT, UNIT, "Sudoku");
    u8g_DrawStr(&u8g, UNIT, 2 * UNIT, "Select level");
    u8g_DrawBox(&u8g, UNIT, menu_cursor, 4, 4);
    u8g_DrawStr(&u8g, 2 * UNIT, 3 * UNIT, "Easy");
    u8g_DrawStr(&u8g, 2 * UNIT, 4 * UNIT, "Medium");
    u8g_DrawStr(&u8g, 2 * UNIT, 5 * UNIT, "Hard");
}

/* Update the menu routine */
void update_menu(void) {

    /* DOWN */
    if (!(PINC & (1 << PC1))) {
        menu_cursor = menu_cursor + UNIT;
        if (menu_cursor > (MENU_CURSOR + 2 * UNIT))
            menu_cursor -= (3 * UNIT);
    }

    /* SELECT the level*/
    if (!(PINC & (1 << PC4))) {
        if (menu_cursor == MENU_CURSOR){
            puzzle = easy;
            fixed_puzzle = easy_fixed;
            puzzle_completed = easy_completed;
        }
        if (menu_cursor == (MENU_CURSOR + UNIT)){
            puzzle = medium;
            fixed_puzzle = medium_fixed;
            puzzle_completed = medium_completed;
        }
        if (menu_cursor == (MENU_CURSOR + 2 * UNIT)){
            puzzle = hard;
            fixed_puzzle = hard_fixed;
            puzzle_completed = hard_completed;
        }
        /* Change the state */
        current_state = UNLOCKED;
    }
}

/* Draw the grid */
void draw_grid(void) {
    uint8_t i;

    for (i = 0; i <= SIZE; i++) {
        u8g_DrawBox(&u8g, 0, i * (BOX_HEIGHT + 1), TABLE_WIDTH, 1);
        u8g_DrawBox(&u8g, i * (BOX_WIDTH + 1), 0, 1, TABLE_HEIGHT);
    }

    u8g_DrawBox(&u8g, 3 * (BOX_WIDTH + 1) - 1, 3 * (BOX_HEIGHT + 1) - 1, 3, 3);
    u8g_DrawBox(&u8g, 6 * (BOX_WIDTH + 1) - 1, 3 * (BOX_HEIGHT + 1) - 1, 3, 3);
    u8g_DrawBox(&u8g, 3 * (BOX_WIDTH + 1) - 1, 6 * (BOX_HEIGHT + 1) - 1, 3, 3);
    u8g_DrawBox(&u8g, 6 * (BOX_WIDTH + 1) - 1, 6 * (BOX_HEIGHT + 1) - 1, 3, 3);
}

/* Draw the values */
void draw_sudoku(void) {

    uint8_t i, j;
    for (i = 0; i < SIZE; i++)
        for (j = 0; j < SIZE; j++) {
        	if (fixed_puzzle[j * SIZE + i] != 0)
            	u8g_DrawBox(&u8g, i * (BOX_WIDTH + 1) + 1, j * (BOX_HEIGHT + 1) + 3, 1, 2);

            if (puzzle[j * SIZE + i] != 0) {
                u8g_DrawStr(&u8g, i * 9 + 3, (j + 1) * 7, digits[puzzle[j * SIZE + i]]);
            }
    }
}

/* Draw the current position of the cursor */
void draw_cursor(void) {

	u8g_DrawBox(&u8g, (current_pos.x + 1) * (BOX_WIDTH + 1) - 1, current_pos.y * (BOX_HEIGHT + 1), 1, BOX_HEIGHT + 1);
}

/* Draw the game */
void draw_game(void) {

    uint8_t REF = 86;

	u8g_SetDefaultForegroundColor(&u8g);
    u8g_SetFont(&u8g, u8g_font_04b_03b);
    draw_grid();
    draw_sudoku();
    draw_cursor();
    if (current_state == UNLOCKED) {
    	u8g_DrawStr(&u8g, REF, UNIT, "UNLOCKED");
    }
    else
    	u8g_DrawStr(&u8g, REF, UNIT, "LOCKED");

    /* Display also state and other info */
    u8g_DrawStr(&u8g, REF, 2 * UNIT, digits[puzzle_completed/10]);
    u8g_DrawStr(&u8g, REF + UNIT/2, 2 * UNIT, digits[puzzle_completed%10]);
    u8g_DrawStr(&u8g, REF + UNIT, 2 * UNIT , "/");
    u8g_DrawStr(&u8g, REF + 3*UNIT/2, 2 * UNIT , "81");

    u8g_DrawStr(&u8g, REF, 5 * UNIT, "L-8");
    u8g_DrawStr(&u8g, REF + UNIT * 2, 5 * UNIT, "R-2");
    u8g_DrawStr(&u8g, REF + UNIT, 4 * UNIT, "U-1");
    u8g_DrawStr(&u8g, REF + UNIT, 6 * UNIT, "D-4");

}

/* Update the cursor when UNLOCKED */
void update_cursor(void) {

	/* UP */
	if (!(PINC & (1 << PC2))) {
		(current_pos.y > 0)?(current_pos.y -= 1):(current_pos.y = SIZE - 1);
	}

	/* DOWN */
	if (!(PINC & (1 << PC1))) {
		current_pos.y = (current_pos.y + 1) % SIZE;
	}

	/* RIGHT */
	if (!(PINC & (1 << PC3))) {
		current_pos.x = (current_pos.x + 1) % SIZE;
	}

	/* LEFT */
	if (!(PINC & (1 << PC0))) {
		(current_pos.x > 0)?(current_pos.x -= 1):(current_pos.x = SIZE - 1);
	}

	/* SELECT */
	if (!(PINC & (1 << PC4))) {
        /* If the users tries to select an invalid box, light the user LED */
		if (fixed_puzzle[current_pos.x + current_pos.y * SIZE] != 0) {
			PORTD ^= (1 << PD7);
			_delay_ms(DELAY);
			PORTD ^= (1 << PD7);
		}
		else 
            /* Valid box select, change state */
			current_state = ~current_state;
	}
}

/* Check if value is valid */
uint8_t check_valid(uint8_t value) {

	uint8_t i, j;
	/* Check line */
	for (i = 0; i < SIZE; i++) {
		if (puzzle[current_pos.y * SIZE + i] == 0)
			continue;
		if ((i != current_pos.x) && (puzzle[current_pos.y * SIZE + i] == value))
			return NOT_OK;
	}

	/* Check column */
	for (i = 0; i < SIZE; i++) {
		if (puzzle[i * SIZE + current_pos.x] == 0)
			continue;
		if ((i != current_pos.y) && (puzzle[i * SIZE + current_pos.x] == value))
			return NOT_OK;
	}


	/* Check box */
	for (i = (current_pos.x - current_pos.x % 3); i < (current_pos.x - current_pos.x % 3 + 3); i++)
		for (j = (current_pos.y - current_pos.y % 3); j < (current_pos.y - current_pos.y % 3 + 3); j++) {
			if (puzzle[j * SIZE + i] == 0)
				continue;
			if ((i != current_pos.x) && (current_pos.y != j) && (puzzle[j * SIZE + i] == value))
				return NOT_OK;
	}

	return OK;
}

/* Update the value */
void update_value(void) {

	uint8_t rc;
	uint8_t value = puzzle[current_pos.y * SIZE + current_pos.x];
	uint8_t init_value = value;

	/* SELECT */
	if (!(PINC & (1 << PC4))) {
		current_state = ~current_state;
		return;
	}

	/* UP - 1 */
	if (!(PINC & (1 << PC2))) {
		if (value % 2 == 0)
			value += 1;
		else
			value -= 1;
	}

	/* DOWN - 4 */
	if (!(PINC & (1 << PC1))) {
		if (value < 8) {
			if (value & (1 << 2))
				value -= 4;
			else
				value += 4;
		}
	}

	/* RIGHT - 2 */
	if (!(PINC & (1 << PC3))) {
		if (value < 8) {
			if (value & (1 << 1))
				value -= 2;
			else
				value += 2;
		}
	}

	/* LEFT - 8 */
	if (!(PINC & (1 << PC0))) {
		if (value > 7)
			value -= 8;
		else if (value < 2)
			value += 8;
	}

	if (value == init_value)
		return;

	rc = check_valid(value);
	if (rc == OK) {
		puzzle[current_pos.y * SIZE + current_pos.x] = value;
		if ((init_value == 0) && (value != 0))
			puzzle_completed++;
		else if ((init_value != 0) && (value == 0))
			puzzle_completed--;
	}
	else {
        /* If value is not valid, signal using the user LED */
		PORTD ^= (1 << PD7);
		_delay_ms(DELAY);
		PORTD ^= (1 << PD7);
	}
}

/* Procedure called by the interrupt routine */
void check_input(void) {

    /* If the user is selecting the level, update menu cursor */
    if (current_state == SELECT_LEVEL)
        update_menu();
    /* If the user is selecting the box, update cursor */
    else if (current_state == UNLOCKED)
        update_cursor();
    else
    /* If the user is selecting the value, update the value */
        update_value();
} 

/* Interrupt routine */
ISR(PCINT2_vect) {
    check_input();
    if (current_state == SELECT_LEVEL)
        draw_level_select();
    else 
        draw_game();

    _delay_ms(DELAY / 5);
}


int main(void) {

    init_lcd();
    init_buttons();
    init_game();
    sei();

    while (1) {
    	while (1) {

        	u8g_FirstPage(&u8g);
        	do {
            	draw_level_select();
        	} while (u8g_NextPage(&u8g));

        	if (current_state != SELECT_LEVEL)
            	break;
    	}

    	while (1) {

        	u8g_FirstPage(&u8g);
        	do {
            	draw_game();
        	} while (u8g_NextPage(&u8g));
    	
        	if (puzzle_completed == (SIZE * SIZE)) {
            
                u8g_FirstPage(&u8g);
        		do {
    				u8g_SetDefaultForegroundColor(&u8g);
    				u8g_SetFont(&u8g, u8g_font_6x10);
    				u8g_DrawStr(&u8g, 32, 32, "Congrats!");
        		} while (u8g_NextPage(&u8g));
                _delay_ms(5000);


            	init_game();
            	memcpy(puzzle, fixed_puzzle, sizeof(puzzle[0]) * SIZE * SIZE);
            	break;
        	}
    	}
    }

    return 0;
}